@@ -24,14 +24,28 @@ if [ "$#" -ne 1 ]; then
   exit 1
 fi
 
-if [ "$(uname -m)" != "x86_64" ]; then
-  echo "GCS testbench won't install on non-x86 architecture"
-  exit 0
-fi
+case "$(uname -m)" in
+  aarch64|arm64|x86_64)
+    : # OK
+    ;;
+  *)
+    echo "GCS testbench is installed only on x86 or arm architectures: $(uname -m)"
+    exit 0
+    ;;
+esac
+
+case "$(uname -s)-$(uname -m)" in
+  Darwin-arm64)
+    # Workaround for https://github.com/grpc/grpc/issues/28387 .
+    # Build grpcio instead of using wheel.
+    # storage-testbench 0.27.0 pins grpcio to 1.46.1.
+    ${PYTHON:-python3} -m pip install --no-binary :all: "grpcio==1.46.1"
+    ;;
+esac
 
 version=$1
 if [[ "${version}" -eq "default" ]]; then
-  version="v0.16.0"
+  version="v0.27.0"
 fi
 
 ${PYTHON:-python3} -m pip install \
