@@ -35,6 +35,7 @@
 #include "arrow/compute/kernel.h"
 #include "arrow/compute/registry.h"
 #include "arrow/memory_pool.h"
+#include "arrow/record_batch.h"
 #include "arrow/scalar.h"
 #include "arrow/status.h"
 #include "arrow/type.h"
@@ -43,7 +44,6 @@
 #include "arrow/util/checked_cast.h"
 #include "arrow/util/cpu_info.h"
 #include "arrow/util/logging.h"
-#include "arrow/util/make_unique.h"
 
 namespace arrow {
 
@@ -56,6 +56,68 @@ using ::arrow::internal::BitmapEquals;
 using ::arrow::internal::CopyBitmap;
 using ::arrow::internal::CountSetBits;
 
+TEST(ExecBatch, SliceBasics) {
+  int64_t length = 4, cut_length = 2, left_length = length - cut_length;
+  ExecBatch batch{{Int32Scalar(0), ArrayFromJSON(utf8(), R"(["a", "b", "c", "d"])"),
+                   ChunkedArrayFromJSON(float64(), {"[1.1]", "[2.2]", "[3.3]", "[4.4]"})},
+                  length};
+  std::vector<ExecBatch> expected_sliced{
+      {{Int32Scalar(0), ArrayFromJSON(utf8(), R"(["a", "b"])"),
+        ChunkedArrayFromJSON(float64(), {"[1.1]", "[2.2]"})},
+       cut_length},
+      {{Int32Scalar(0), ArrayFromJSON(utf8(), R"(["c", "d"])"),
+        ChunkedArrayFromJSON(float64(), {"[3.3]", "[4.4]"})},
+       left_length}};
+  std::vector<ExecBatch> actual_sliced = {batch.Slice(0, cut_length),
+                                          batch.Slice(cut_length, left_length)};
+  for (size_t i = 0; i < expected_sliced.size(); i++) {
+    ASSERT_EQ(expected_sliced[i].length, actual_sliced[i].length);
+    ASSERT_EQ(expected_sliced[i].values.size(), actual_sliced[i].values.size());
+    for (size_t j = 0; j < expected_sliced[i].values.size(); j++) {
+      AssertDatumsEqual(expected_sliced[i].values[j], actual_sliced[i].values[j]);
+    }
+    ASSERT_EQ(expected_sliced[i].ToString(), actual_sliced[i].ToString());
+  }
+}
+
+TEST(ExecBatch, ToRecordBatch) {
+  auto i32_array = ArrayFromJSON(int32(), "[0, 1, 2]");
+  auto utf8_array = ArrayFromJSON(utf8(), R"(["a", "b", "c"])");
+  ExecBatch exec_batch({Datum(i32_array), Datum(utf8_array)}, 3);
+
+  auto right_schema = schema({field("a", int32()), field("b", utf8())});
+  ASSERT_OK_AND_ASSIGN(auto right_record_batch, exec_batch.ToRecordBatch(right_schema));
+  ASSERT_OK(right_record_batch->ValidateFull());
+  auto expected_batch = RecordBatchFromJSON(right_schema, R"([
+      {"a": 0, "b": "a"},
+      {"a": 1, "b": "b"},
+      {"a": 2, "b": "c"}
+      ])");
+  AssertBatchesEqual(*right_record_batch, *expected_batch);
+
+  // With a scalar column
+  auto utf8_scalar = ScalarFromJSON(utf8(), R"("z")");
+  exec_batch = ExecBatch({Datum(i32_array), Datum(utf8_scalar)}, 3);
+  ASSERT_OK_AND_ASSIGN(right_record_batch, exec_batch.ToRecordBatch(right_schema));
+  ASSERT_OK(right_record_batch->ValidateFull());
+  expected_batch = RecordBatchFromJSON(right_schema, R"([
+      {"a": 0, "b": "z"},
+      {"a": 1, "b": "z"},
+      {"a": 2, "b": "z"}
+      ])");
+  AssertBatchesEqual(*right_record_batch, *expected_batch);
+
+  // Wrong number of fields in schema
+  auto reject_schema =
+      schema({field("a", int32()), field("b", utf8()), field("c", float64())});
+  ASSERT_RAISES(Invalid, exec_batch.ToRecordBatch(reject_schema));
+
+  // Wrong-kind exec batch (not really valid, but test it here anyway)
+  ExecBatch miskinded_batch({Datum()}, 0);
+  auto null_schema = schema({field("a", null())});
+  ASSERT_RAISES(TypeError, miskinded_batch.ToRecordBatch(null_schema));
+}
+
 TEST(ExecContext, BasicWorkings) {
   {
     ExecContext ctx;
@@ -90,13 +152,20 @@ TEST(SelectionVector, Basics) {
   ASSERT_EQ(3, sel_vector->indices()[1]);
 }
 
+void AssertValidityZeroExtraBits(const uint8_t* data, int64_t length, int64_t offset) {
+  const int64_t bit_extent = ((offset + length + 7) / 8) * 8;
+  for (int64_t i = offset + length; i < bit_extent; ++i) {
+    EXPECT_FALSE(bit_util::GetBit(data, i)) << i;
+  }
+}
+
+void AssertValidityZeroExtraBits(const ArraySpan& arr) {
+  return AssertValidityZeroExtraBits(arr.buffers[0].data, arr.length, arr.offset);
+}
+
 void AssertValidityZeroExtraBits(const ArrayData& arr) {
   const Buffer& buf = *arr.buffers[0];
-
-  const int64_t bit_extent = ((arr.offset + arr.length + 7) / 8) * 8;
-  for (int64_t i = arr.offset + arr.length; i < bit_extent; ++i) {
-    EXPECT_FALSE(bit_util::GetBit(buf.data(), i)) << i;
-  }
+  return AssertValidityZeroExtraBits(buf.data(), arr.length, arr.offset);
 }
 
 class TestComputeInternals : public ::testing::Test {
@@ -137,6 +206,9 @@ class TestComputeInternals : public ::testing::Test {
   std::unique_ptr<random::RandomArrayGenerator> rng_;
 };
 
+// ----------------------------------------------------------------------
+// Test PropagateNulls
+
 class TestPropagateNulls : public TestComputeInternals {};
 
 TEST_F(TestPropagateNulls, UnknownNullCountWithNullsZeroCopies) {
@@ -149,7 +221,7 @@ TEST_F(TestPropagateNulls, UnknownNullCountWithNullsZeroCopies) {
   ArrayData input(boolean(), length, {nulls, nullptr}, kUnknownNullCount);
 
   ExecBatch batch({input}, length);
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
   ASSERT_EQ(nulls.get(), output.buffers[0].get());
   ASSERT_EQ(kUnknownNullCount, output.null_count);
   ASSERT_EQ(9, output.GetNullCount());
@@ -164,7 +236,7 @@ TEST_F(TestPropagateNulls, UnknownNullCountWithoutNulls) {
   ArrayData input(boolean(), length, {nulls, nullptr}, kUnknownNullCount);
 
   ExecBatch batch({input}, length);
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
   EXPECT_EQ(-1, output.null_count);
   EXPECT_EQ(nulls.get(), output.buffers[0].get());
 }
@@ -185,7 +257,7 @@ TEST_F(TestPropagateNulls, SetAllNulls) {
     ArrayData output(boolean(), length, buffers);
 
     ExecBatch batch(values, length);
-    ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+    ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
 
     if (preallocate) {
       // Ensure that buffer object the same when we pass in preallocated memory
@@ -228,7 +300,7 @@ TEST_F(TestPropagateNulls, SetAllNulls) {
     // null-scalar earlier in the batch
     ArrayData output(boolean(), length, {nullptr, nullptr});
     ExecBatch batch({MakeNullScalar(boolean()), arr_all_nulls}, length);
-    ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+    ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
     ASSERT_EQ(arr_all_nulls->data()->buffers[0].get(), output.buffers[0].get());
   }
 }
@@ -260,7 +332,7 @@ TEST_F(TestPropagateNulls, SingleValueWithNulls) {
       ASSERT_EQ(0, output.offset);
     }
 
-    ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+    ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
 
     if (!preallocate) {
       const Buffer* parent_buf = arr->data()->buffers[0].get();
@@ -308,14 +380,14 @@ TEST_F(TestPropagateNulls, ZeroCopyWhenZeroNullsOnOneInput) {
 
   ArrayData output(boolean(), length, {nullptr, nullptr});
   ExecBatch batch({some_nulls, no_nulls}, length);
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
   ASSERT_EQ(nulls.get(), output.buffers[0].get());
   ASSERT_EQ(9, output.null_count);
 
   // Flip order of args
   output = ArrayData(boolean(), length, {nullptr, nullptr});
   batch.values = {no_nulls, no_nulls, some_nulls};
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
   ASSERT_EQ(nulls.get(), output.buffers[0].get());
   ASSERT_EQ(9, output.null_count);
 
@@ -324,7 +396,7 @@ TEST_F(TestPropagateNulls, ZeroCopyWhenZeroNullsOnOneInput) {
   auto preallocated_mem = std::make_shared<MutableBuffer>(bitmap_data, 2);
   output.null_count = kUnknownNullCount;
   output.buffers[0] = preallocated_mem;
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
 
   ASSERT_EQ(preallocated_mem.get(), output.buffers[0].get());
   ASSERT_EQ(9, output.null_count);
@@ -366,7 +438,7 @@ TEST_F(TestPropagateNulls, IntersectsNulls) {
     ArrayData output(boolean(), length, {nulls, nullptr});
     output.offset = output_offset;
 
-    ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+    ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
 
     // Preallocated memory used
     if (preallocate) {
@@ -405,48 +477,286 @@ TEST_F(TestPropagateNulls, NullOutputTypeNoop) {
   ExecBatch batch({rng_->Boolean(100, 0.5, 0.5)}, length);
 
   ArrayData output(null(), length, {nullptr});
-  ASSERT_OK(PropagateNulls(ctx_.get(), batch, &output));
+  ASSERT_OK(PropagateNulls(ctx_.get(), ExecSpan(batch), &output));
   ASSERT_EQ(nullptr, output.buffers[0]);
 }
 
 // ----------------------------------------------------------------------
-// ExecBatchIterator
+// Test PropagateNullsSpans (new span-based implementation). Some of
+// the tests above had to be rewritten because the span-based
+// implementation does not deal with zero-copy optimizations right now
 
-class TestExecBatchIterator : public TestComputeInternals {
- public:
-  void SetupIterator(std::vector<Datum> args,
-                     int64_t max_chunksize = kDefaultMaxChunksize) {
-    ASSERT_OK_AND_ASSIGN(iterator_,
-                         ExecBatchIterator::Make(std::move(args), max_chunksize));
+class TestPropagateNullsSpans : public TestComputeInternals {};
+
+TEST_F(TestPropagateNullsSpans, UnknownNullCountWithNullsZeroCopies) {
+  const int64_t length = 16;
+
+  const uint8_t validity_bitmap[8] = {254, 0, 0, 0, 0, 0, 0, 0};
+  auto nulls = std::make_shared<Buffer>(validity_bitmap, 8);
+  auto ty = boolean();
+  ArrayData input(ty, length, {nulls, nullptr}, kUnknownNullCount);
+
+  uint8_t validity_bitmap2[8] = {0, 0, 0, 0, 0, 0, 0, 0};
+  auto nulls2 = std::make_shared<Buffer>(validity_bitmap2, 8);
+  ArraySpan output(ty.get(), length);
+  output.buffers[0].data = validity_bitmap2;
+  output.buffers[0].size = 0;
+
+  ExecSpan span(ExecBatch({input}, length));
+  PropagateNullsSpans(span, &output);
+  ASSERT_EQ(kUnknownNullCount, output.null_count);
+  ASSERT_EQ(9, output.GetNullCount());
+}
+
+TEST_F(TestPropagateNullsSpans, UnknownNullCountWithoutNulls) {
+  const int64_t length = 16;
+  constexpr uint8_t validity_bitmap[8] = {255, 255, 0, 0, 0, 0, 0, 0};
+  auto nulls = std::make_shared<Buffer>(validity_bitmap, 8);
+
+  auto ty = boolean();
+  ArrayData input(ty, length, {nulls, nullptr}, kUnknownNullCount);
+
+  uint8_t validity_bitmap2[8] = {0, 0, 0, 0, 0, 0, 0, 0};
+  auto nulls2 = std::make_shared<Buffer>(validity_bitmap2, 8);
+  ArraySpan output(ty.get(), length);
+  output.buffers[0].data = validity_bitmap2;
+  output.buffers[0].size = 0;
+
+  ExecSpan span(ExecBatch({input}, length));
+  PropagateNullsSpans(span, &output);
+  ASSERT_EQ(kUnknownNullCount, output.null_count);
+  ASSERT_EQ(0, output.GetNullCount());
+}
+
+TEST_F(TestPropagateNullsSpans, SetAllNulls) {
+  const int64_t length = 16;
+
+  auto CheckSetAllNull = [&](std::vector<Datum> values) {
+    // Make fresh bitmap with all 1's
+    uint8_t bitmap_data[2] = {255, 255};
+    auto buf = std::make_shared<MutableBuffer>(bitmap_data, 2);
+
+    auto ty = boolean();
+    ArraySpan output(ty.get(), length);
+    output.SetBuffer(0, buf);
+
+    ExecSpan span(ExecBatch(values, length));
+    PropagateNullsSpans(span, &output);
+
+    uint8_t expected[2] = {0, 0};
+    ASSERT_EQ(0, std::memcmp(output.buffers[0].data, expected, output.buffers[0].size));
+  };
+
+  // There is a null scalar
+  std::shared_ptr<Scalar> i32_val = std::make_shared<Int32Scalar>(3);
+  std::vector<Datum> vals = {i32_val, MakeNullScalar(boolean())};
+  CheckSetAllNull(vals);
+
+  const double true_prob = 0.5;
+  vals[0] = rng_->Boolean(length, true_prob);
+  CheckSetAllNull(vals);
+
+  auto arr_all_nulls = rng_->Boolean(length, true_prob, /*null_probability=*/1);
+
+  // One value is all null
+  vals = {rng_->Boolean(length, true_prob, /*null_probability=*/0.5), arr_all_nulls};
+  CheckSetAllNull(vals);
+
+  // A value is NullType
+  std::shared_ptr<Array> null_arr = std::make_shared<NullArray>(length);
+  vals = {rng_->Boolean(length, true_prob), null_arr};
+  CheckSetAllNull(vals);
+}
+
+TEST_F(TestPropagateNullsSpans, SingleValueWithNulls) {
+  // Input offset is non-zero (0 mod 8 and nonzero mod 8 cases)
+  const int64_t length = 100;
+  auto arr = rng_->Boolean(length, 0.5, /*null_probability=*/0.5);
+
+  auto CheckSliced = [&](int64_t offset, int64_t out_offset = 0) {
+    // Unaligned bitmap, zero copy not possible
+    auto sliced = arr->Slice(offset);
+    std::vector<Datum> vals = {sliced};
+
+    auto ty = boolean();
+    ArraySpan output(ty.get(), vals[0].length());
+    output.offset = out_offset;
+
+    std::shared_ptr<Buffer> preallocated_bitmap;
+    ASSERT_OK_AND_ASSIGN(
+        preallocated_bitmap,
+        AllocateBuffer(bit_util::BytesForBits(sliced->length() + out_offset)));
+    std::memset(preallocated_bitmap->mutable_data(), 0, preallocated_bitmap->size());
+    output.SetBuffer(0, preallocated_bitmap);
+
+    ExecBatch batch(vals, vals[0].length());
+    PropagateNullsSpans(ExecSpan(batch), &output);
+    ASSERT_EQ(arr->Slice(offset)->null_count(), output.GetNullCount());
+    ASSERT_TRUE(BitmapEquals(output.buffers[0].data, output.offset,
+                             sliced->null_bitmap_data(), sliced->offset(),
+                             output.length));
+    AssertValidityZeroExtraBits(output);
+  };
+
+  CheckSliced(8);
+  CheckSliced(7);
+  CheckSliced(8, /*offset=*/4);
+  CheckSliced(7, /*offset=*/4);
+}
+
+TEST_F(TestPropagateNullsSpans, CasesThatUsedToBeZeroCopy) {
+  // ARROW-16576: testing behaviors that used to be zero copy but are
+  // not anymore
+  const int64_t length = 16;
+
+  auto ty = boolean();
+  constexpr uint8_t validity_bitmap[8] = {254, 0, 0, 0, 0, 0, 0, 0};
+  auto nulls = std::make_shared<Buffer>(validity_bitmap, 8);
+
+  ArraySpan some_nulls(ty.get(), length);
+  some_nulls.SetBuffer(0, nulls);
+  some_nulls.null_count = 9;
+
+  ArraySpan no_nulls(ty.get(), length);
+  no_nulls.null_count = 0;
+
+  {
+    uint8_t bitmap_data[2] = {0, 0};
+    auto preallocated_mem = std::make_shared<Buffer>(bitmap_data, 2);
+
+    ArraySpan output(ty.get(), length);
+    output.SetBuffer(0, preallocated_mem);
+    PropagateNullsSpans(ExecSpan({some_nulls, no_nulls}, length), &output);
+    ASSERT_EQ(
+        0, std::memcmp(output.buffers[0].data, validity_bitmap, output.buffers[0].size));
+    ASSERT_EQ(output.buffers[0].owner, &preallocated_mem);
+    ASSERT_EQ(9, output.GetNullCount());
   }
-  void CheckIteration(const std::vector<Datum>& args, int chunksize,
+
+  // Flip order of args
+  {
+    uint8_t bitmap_data[2] = {0, 0};
+    auto preallocated_mem = std::make_shared<Buffer>(bitmap_data, 2);
+
+    ArraySpan output(ty.get(), length);
+    output.SetBuffer(0, preallocated_mem);
+    PropagateNullsSpans(ExecSpan({no_nulls, no_nulls, some_nulls}, length), &output);
+    ASSERT_EQ(
+        0, std::memcmp(output.buffers[0].data, validity_bitmap, output.buffers[0].size));
+    ASSERT_EQ(output.buffers[0].owner, &preallocated_mem);
+    ASSERT_EQ(9, output.GetNullCount());
+  }
+}
+
+TEST_F(TestPropagateNullsSpans, IntersectsNulls) {
+  const int64_t length = 16;
+
+  // 0b01111111 0b11001111
+  constexpr uint8_t bitmap1[8] = {127, 207, 0, 0, 0, 0, 0, 0};
+  auto buffer1 = std::make_shared<Buffer>(bitmap1, 8);
+
+  // 0b11111110 0b01111111
+  constexpr uint8_t bitmap2[8] = {254, 127, 0, 0, 0, 0, 0, 0};
+  auto buffer2 = std::make_shared<Buffer>(bitmap2, 8);
+
+  // 0b11101111 0b11111110
+  constexpr uint8_t bitmap3[8] = {239, 254, 0, 0, 0, 0, 0, 0};
+  auto buffer3 = std::make_shared<Buffer>(bitmap3, 8);
+
+  auto ty = boolean();
+
+  ArraySpan arr1(ty.get(), length);
+  arr1.SetBuffer(0, buffer1);
+
+  ArraySpan arr2(ty.get(), length);
+  arr2.SetBuffer(0, buffer2);
+
+  ArraySpan arr3(ty.get(), length);
+  arr3.SetBuffer(0, buffer3);
+
+  auto CheckCase = [&](std::vector<ExecValue> values, int64_t ex_null_count,
+                       const uint8_t* ex_bitmap, int64_t output_offset = 0) {
+    ExecSpan batch(values, length);
+
+    std::shared_ptr<Buffer> nulls;
+    // Make the buffer one byte bigger so we can have non-zero offsets
+    ASSERT_OK_AND_ASSIGN(nulls, AllocateBuffer(3));
+    std::memset(nulls->mutable_data(), 0, nulls->size());
+
+    ArraySpan output(ty.get(), length);
+    output.SetBuffer(0, nulls);
+    output.offset = output_offset;
+
+    PropagateNullsSpans(batch, &output);
+    ASSERT_EQ(&nulls, output.buffers[0].owner);
+    EXPECT_EQ(kUnknownNullCount, output.null_count);
+    EXPECT_EQ(ex_null_count, output.GetNullCount());
+    ASSERT_TRUE(BitmapEquals(output.buffers[0].data, output_offset, ex_bitmap,
+                             /*ex_offset=*/0, length));
+
+    // Now check that the rest of the bits in out_buffer are still 0
+    AssertValidityZeroExtraBits(output);
+  };
+
+  // 0b01101110 0b01001110
+  uint8_t expected1[2] = {110, 78};
+  CheckCase({arr1, arr2, arr3}, 7, expected1);
+  CheckCase({arr1, arr2, arr3}, 7, expected1, /*output_offset=*/4);
+
+  // 0b01111110 0b01001111
+  uint8_t expected2[2] = {126, 79};
+  CheckCase({arr1, arr2}, 5, expected2, /*output_offset=*/4);
+}
+
+TEST_F(TestPropagateNullsSpans, NullOutputTypeNoop) {
+  // Ensure we leave the buffers alone when the output type is null()
+  // TODO(wesm): is this test useful? Can probably delete
+  const int64_t length = 100;
+  ExecBatch batch({rng_->Boolean(100, 0.5, 0.5)}, length);
+
+  auto ty = null();
+  ArraySpan result(ty.get(), length);
+  PropagateNullsSpans(ExecSpan(batch), &result);
+  ASSERT_EQ(nullptr, result.buffers[0].data);
+}
+
+// ----------------------------------------------------------------------
+// ExecSpanIterator tests
+
+class TestExecSpanIterator : public TestComputeInternals {
+ public:
+  void SetupIterator(const ExecBatch& batch,
+                     int64_t max_chunksize = kDefaultMaxChunksize) {
+    ASSERT_OK(iterator_.Init(batch, max_chunksize));
+  }
+  void CheckIteration(const ExecBatch& input, int chunksize,
                       const std::vector<int>& ex_batch_sizes) {
-    SetupIterator(args, chunksize);
-    ExecBatch batch;
+    SetupIterator(input, chunksize);
+    ExecSpan batch;
     int64_t position = 0;
     for (size_t i = 0; i < ex_batch_sizes.size(); ++i) {
-      ASSERT_EQ(position, iterator_->position());
-      ASSERT_TRUE(iterator_->Next(&batch));
+      ASSERT_EQ(position, iterator_.position());
+      ASSERT_TRUE(iterator_.Next(&batch));
       ASSERT_EQ(ex_batch_sizes[i], batch.length);
 
-      for (size_t j = 0; j < args.size(); ++j) {
-        switch (args[j].kind()) {
+      for (size_t j = 0; j < input.values.size(); ++j) {
+        switch (input[j].kind()) {
           case Datum::SCALAR:
-            ASSERT_TRUE(args[j].scalar()->Equals(batch[j].scalar()));
+            ASSERT_TRUE(input[j].scalar()->Equals(*batch[j].scalar));
             break;
           case Datum::ARRAY:
-            AssertArraysEqual(*args[j].make_array()->Slice(position, batch.length),
-                              *batch[j].make_array());
+            AssertArraysEqual(*input[j].make_array()->Slice(position, batch.length),
+                              *batch[j].array.ToArray());
             break;
           case Datum::CHUNKED_ARRAY: {
-            const ChunkedArray& carr = *args[j].chunked_array();
+            const ChunkedArray& carr = *input[j].chunked_array();
             if (batch.length == 0) {
               ASSERT_EQ(0, carr.length());
             } else {
               auto arg_slice = carr.Slice(position, batch.length);
               // The sliced ChunkedArrays should only ever be 1 chunk
               ASSERT_EQ(1, arg_slice->num_chunks());
-              AssertArraysEqual(*arg_slice->chunk(0), *batch[j].make_array());
+              AssertArraysEqual(*arg_slice->chunk(0), *batch[j].array.ToArray());
             }
           } break;
           default:
@@ -456,143 +766,157 @@ class TestExecBatchIterator : public TestComputeInternals {
       position += ex_batch_sizes[i];
     }
     // Ensure that the iterator is exhausted
-    ASSERT_FALSE(iterator_->Next(&batch));
+    ASSERT_FALSE(iterator_.Next(&batch));
 
-    ASSERT_EQ(iterator_->length(), iterator_->position());
+    ASSERT_EQ(iterator_.length(), iterator_.position());
   }
 
  protected:
-  std::unique_ptr<ExecBatchIterator> iterator_;
+  ExecSpanIterator iterator_;
 };
 
-TEST_F(TestExecBatchIterator, Basics) {
+TEST_F(TestExecSpanIterator, Basics) {
   const int64_t length = 100;
 
-  // Simple case with a single chunk
-  std::vector<Datum> args = {Datum(GetInt32Array(length)), Datum(GetFloat64Array(length)),
-                             Datum(std::make_shared<Int32Scalar>(3))};
-  SetupIterator(args);
+  ExecBatch input;
+  input.length = 100;
 
-  ExecBatch batch;
-  ASSERT_TRUE(iterator_->Next(&batch));
+  // Simple case with a single chunk
+  input.values = {Datum(GetInt32Array(length)), Datum(GetFloat64Array(length)),
+                  Datum(std::make_shared<Int32Scalar>(3))};
+  SetupIterator(input);
+
+  ExecSpan batch;
+  ASSERT_TRUE(iterator_.Next(&batch));
   ASSERT_EQ(3, batch.values.size());
   ASSERT_EQ(3, batch.num_values());
   ASSERT_EQ(length, batch.length);
 
-  std::vector<ValueDescr> descrs = batch.GetDescriptors();
-  ASSERT_EQ(ValueDescr::Array(int32()), descrs[0]);
-  ASSERT_EQ(ValueDescr::Array(float64()), descrs[1]);
-  ASSERT_EQ(ValueDescr::Scalar(int32()), descrs[2]);
+  AssertArraysEqual(*input[0].make_array(), *batch[0].array.ToArray());
+  AssertArraysEqual(*input[1].make_array(), *batch[1].array.ToArray());
+  ASSERT_TRUE(input[2].scalar()->Equals(*batch[2].scalar));
 
-  AssertArraysEqual(*args[0].make_array(), *batch[0].make_array());
-  AssertArraysEqual(*args[1].make_array(), *batch[1].make_array());
-  ASSERT_TRUE(args[2].scalar()->Equals(batch[2].scalar()));
-
-  ASSERT_EQ(length, iterator_->position());
-  ASSERT_FALSE(iterator_->Next(&batch));
+  ASSERT_EQ(length, iterator_.position());
+  ASSERT_FALSE(iterator_.Next(&batch));
 
   // Split into chunks of size 16
-  CheckIteration(args, /*chunksize=*/16, {16, 16, 16, 16, 16, 16, 4});
+  CheckIteration(input, /*chunksize=*/16, {16, 16, 16, 16, 16, 16, 4});
 }
 
-TEST_F(TestExecBatchIterator, InputValidation) {
-  std::vector<Datum> args = {Datum(GetInt32Array(10)), Datum(GetInt32Array(9))};
-  ASSERT_RAISES(Invalid, ExecBatchIterator::Make(args));
+TEST_F(TestExecSpanIterator, InputValidation) {
+  ExecSpanIterator iterator;
 
-  args = {Datum(GetInt32Array(9)), Datum(GetInt32Array(10))};
-  ASSERT_RAISES(Invalid, ExecBatchIterator::Make(args));
+  ExecBatch batch({Datum(GetInt32Array(10)), Datum(GetInt32Array(9))}, 10);
+  ASSERT_RAISES(Invalid, iterator.Init(batch));
 
-  args = {Datum(GetInt32Array(10))};
-  ASSERT_OK_AND_ASSIGN(auto iterator, ExecBatchIterator::Make(args));
-  ASSERT_EQ(10, iterator->max_chunksize());
+  batch.values = {Datum(GetInt32Array(9)), Datum(GetInt32Array(10))};
+  ASSERT_RAISES(Invalid, iterator.Init(batch));
+
+  batch.values = {Datum(GetInt32Array(10))};
+  ASSERT_OK(iterator.Init(batch));
 }
 
-TEST_F(TestExecBatchIterator, ChunkedArrays) {
-  std::vector<Datum> args = {Datum(GetInt32Chunked({0, 20, 10})),
-                             Datum(GetInt32Chunked({15, 15})), Datum(GetInt32Array(30)),
-                             Datum(std::make_shared<Int32Scalar>(5)),
-                             Datum(MakeNullScalar(boolean()))};
+TEST_F(TestExecSpanIterator, ChunkedArrays) {
+  ExecBatch batch({Datum(GetInt32Chunked({0, 20, 10})), Datum(GetInt32Chunked({15, 15})),
+                   Datum(GetInt32Array(30)), Datum(std::make_shared<Int32Scalar>(5)),
+                   Datum(MakeNullScalar(boolean()))},
+                  30);
 
-  CheckIteration(args, /*chunksize=*/10, {10, 5, 5, 10});
-  CheckIteration(args, /*chunksize=*/20, {15, 5, 10});
-  CheckIteration(args, /*chunksize=*/30, {15, 5, 10});
+  CheckIteration(batch, /*chunksize=*/10, {10, 5, 5, 10});
+  CheckIteration(batch, /*chunksize=*/20, {15, 5, 10});
+  CheckIteration(batch, /*chunksize=*/30, {15, 5, 10});
 }
 
-TEST_F(TestExecBatchIterator, ZeroLengthInputs) {
-  auto carr = std::shared_ptr<ChunkedArray>(new ChunkedArray({}, int32()));
+TEST_F(TestExecSpanIterator, ZeroLengthInputs) {
+  auto carr = std::make_shared<ChunkedArray>(ArrayVector{}, int32());
 
-  auto CheckArgs = [&](const std::vector<Datum>& args) {
-    auto iterator = ExecBatchIterator::Make(args).ValueOrDie();
-    ExecBatch batch;
-    ASSERT_FALSE(iterator->Next(&batch));
+  auto CheckArgs = [&](const ExecBatch& batch) {
+    ExecSpanIterator iterator;
+    ASSERT_OK(iterator.Init(batch));
+    ExecSpan iter_span;
+    ASSERT_TRUE(iterator.Next(&iter_span));
+    ASSERT_EQ(0, iter_span.length);
+    ASSERT_FALSE(iterator.Next(&iter_span));
   };
 
+  ExecBatch input;
+  input.length = 0;
+
   // Zero-length ChunkedArray with zero chunks
-  std::vector<Datum> args = {Datum(carr)};
-  CheckArgs(args);
+  input.values = {Datum(carr)};
+  CheckArgs(input);
 
   // Zero-length array
-  args = {Datum(GetInt32Array(0))};
-  CheckArgs(args);
+  input.values = {Datum(GetInt32Array(0))};
+  CheckArgs(input);
 
   // ChunkedArray with single empty chunk
-  args = {Datum(GetInt32Chunked({0}))};
-  CheckArgs(args);
+  input.values = {Datum(GetInt32Chunked({0}))};
+  CheckArgs(input);
 }
 
 // ----------------------------------------------------------------------
 // Scalar function execution
 
-Status ExecCopy(KernelContext*, const ExecBatch& batch, Datum* out) {
+Status ExecCopyArrayData(KernelContext*, const ExecSpan& batch, ExecResult* out) {
   DCHECK_EQ(1, batch.num_values());
-  const auto& type = checked_cast<const FixedWidthType&>(*batch[0].type());
-  int value_size = type.bit_width() / 8;
+  int value_size = batch[0].type()->byte_width();
 
-  const ArrayData& arg0 = *batch[0].array();
-  ArrayData* out_arr = out->mutable_array();
+  const ArraySpan& arg0 = batch[0].array;
+  ArrayData* out_arr = out->array_data().get();
   uint8_t* dst = out_arr->buffers[1]->mutable_data() + out_arr->offset * value_size;
-  const uint8_t* src = arg0.buffers[1]->data() + arg0.offset * value_size;
+  const uint8_t* src = arg0.buffers[1].data + arg0.offset * value_size;
   std::memcpy(dst, src, batch.length * value_size);
   return Status::OK();
 }
 
-Status ExecComputedBitmap(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+Status ExecCopyArraySpan(KernelContext*, const ExecSpan& batch, ExecResult* out) {
+  DCHECK_EQ(1, batch.num_values());
+  int value_size = batch[0].type()->byte_width();
+  const ArraySpan& arg0 = batch[0].array;
+  ArraySpan* out_arr = out->array_span();
+  uint8_t* dst = out_arr->buffers[1].data + out_arr->offset * value_size;
+  const uint8_t* src = arg0.buffers[1].data + arg0.offset * value_size;
+  std::memcpy(dst, src, batch.length * value_size);
+  return Status::OK();
+}
+
+Status ExecComputedBitmap(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
   // Propagate nulls not used. Check that the out bitmap isn't the same already
   // as the input bitmap
-  const ArrayData& arg0 = *batch[0].array();
-  ArrayData* out_arr = out->mutable_array();
-
-  if (CountSetBits(arg0.buffers[0]->data(), arg0.offset, batch.length) > 0) {
+  const ArraySpan& arg0 = batch[0].array;
+  ArraySpan* out_arr = out->array_span();
+  if (CountSetBits(arg0.buffers[0].data, arg0.offset, batch.length) > 0) {
     // Check that the bitmap has not been already copied over
-    DCHECK(!BitmapEquals(arg0.buffers[0]->data(), arg0.offset,
-                         out_arr->buffers[0]->data(), out_arr->offset, batch.length));
+    DCHECK(!BitmapEquals(arg0.buffers[0].data, arg0.offset, out_arr->buffers[0].data,
+                         out_arr->offset, batch.length));
   }
 
-  CopyBitmap(arg0.buffers[0]->data(), arg0.offset, batch.length,
-             out_arr->buffers[0]->mutable_data(), out_arr->offset);
-  return ExecCopy(ctx, batch, out);
+  CopyBitmap(arg0.buffers[0].data, arg0.offset, batch.length, out_arr->buffers[0].data,
+             out_arr->offset);
+  return ExecCopyArraySpan(ctx, batch, out);
 }
 
-Status ExecNoPreallocatedData(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+Status ExecNoPreallocatedData(KernelContext* ctx, const ExecSpan& batch,
+                              ExecResult* out) {
   // Validity preallocated, but not the data
-  ArrayData* out_arr = out->mutable_array();
+  ArrayData* out_arr = out->array_data().get();
   DCHECK_EQ(0, out_arr->offset);
-  const auto& type = checked_cast<const FixedWidthType&>(*batch[0].type());
-  int value_size = type.bit_width() / 8;
+  int value_size = batch[0].type()->byte_width();
   Status s = (ctx->Allocate(out_arr->length * value_size).Value(&out_arr->buffers[1]));
   DCHECK_OK(s);
-  return ExecCopy(ctx, batch, out);
+  return ExecCopyArrayData(ctx, batch, out);
 }
 
-Status ExecNoPreallocatedAnything(KernelContext* ctx, const ExecBatch& batch,
-                                  Datum* out) {
+Status ExecNoPreallocatedAnything(KernelContext* ctx, const ExecSpan& batch,
+                                  ExecResult* out) {
   // Neither validity nor data preallocated
-  ArrayData* out_arr = out->mutable_array();
+  ArrayData* out_arr = out->array_data().get();
   DCHECK_EQ(0, out_arr->offset);
   Status s = (ctx->AllocateBitmap(out_arr->length).Value(&out_arr->buffers[0]));
   DCHECK_OK(s);
-  const ArrayData& arg0 = *batch[0].array();
-  CopyBitmap(arg0.buffers[0]->data(), arg0.offset, batch.length,
+  const ArraySpan& arg0 = batch[0].array;
+  CopyBitmap(arg0.buffers[0].data, arg0.offset, batch.length,
              out_arr->buffers[0]->mutable_data(), /*offset=*/0);
 
   // Reuse the kernel that allocates the data
@@ -621,7 +945,7 @@ class ExampleOptionsType : public FunctionOptionsType {
   }
   std::unique_ptr<FunctionOptions> Copy(const FunctionOptions& options) const override {
     const auto& opts = static_cast<const ExampleOptions&>(options);
-    return arrow::internal::make_unique<ExampleOptions>(opts.value);
+    return std::make_unique<ExampleOptions>(opts.value);
   }
 };
 ExampleOptions::ExampleOptions(std::shared_ptr<Scalar> value)
@@ -635,28 +959,31 @@ struct ExampleState : public KernelState {
 Result<std::unique_ptr<KernelState>> InitStateful(KernelContext*,
                                                   const KernelInitArgs& args) {
   auto func_options = static_cast<const ExampleOptions*>(args.options);
-  return std::unique_ptr<KernelState>(new ExampleState{func_options->value});
+  return std::make_unique<ExampleState>(func_options->value);
 }
 
-Status ExecStateful(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
+Status ExecStateful(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
   // We take the value from the state and multiply the data in batch[0] with it
   ExampleState* state = static_cast<ExampleState*>(ctx->state());
   int32_t multiplier = checked_cast<const Int32Scalar&>(*state->value).value;
 
-  const ArrayData& arg0 = *batch[0].array();
-  ArrayData* out_arr = out->mutable_array();
+  const ArraySpan& arg0 = batch[0].array;
+  ArraySpan* out_arr = out->array_span();
   const int32_t* arg0_data = arg0.GetValues<int32_t>(1);
-  int32_t* dst = out_arr->GetMutableValues<int32_t>(1);
+  int32_t* dst = out_arr->GetValues<int32_t>(1);
   for (int64_t i = 0; i < arg0.length; ++i) {
     dst[i] = arg0_data[i] * multiplier;
   }
   return Status::OK();
 }
 
-Status ExecAddInt32(KernelContext* ctx, const ExecBatch& batch, Datum* out) {
-  const Int32Scalar& arg0 = batch[0].scalar_as<Int32Scalar>();
-  const Int32Scalar& arg1 = batch[1].scalar_as<Int32Scalar>();
-  out->value = std::make_shared<Int32Scalar>(arg0.value + arg1.value);
+Status ExecAddInt32(KernelContext* ctx, const ExecSpan& batch, ExecResult* out) {
+  const int32_t* left_data = batch[0].array.GetValues<int32_t>(1);
+  const int32_t* right_data = batch[1].array.GetValues<int32_t>(1);
+  int32_t* out_data = out->array_span()->GetValues<int32_t>(1);
+  for (int64_t i = 0; i < batch.length; ++i) {
+    *out_data++ = *left_data++ + *right_data++;
+  }
   return Status::OK();
 }
 
@@ -681,19 +1008,19 @@ class TestCallScalarFunction : public TestComputeInternals {
 
     // This function simply copies memory from the input argument into the
     // (preallocated) output
-    auto func =
-        std::make_shared<ScalarFunction>("test_copy", Arity::Unary(), /*doc=*/nullptr);
+    auto func = std::make_shared<ScalarFunction>("test_copy", Arity::Unary(),
+                                                 /*doc=*/FunctionDoc::Empty());
 
     // Add a few kernels. Our implementation only accepts arrays
-    ASSERT_OK(func->AddKernel({InputType::Array(uint8())}, uint8(), ExecCopy));
-    ASSERT_OK(func->AddKernel({InputType::Array(int32())}, int32(), ExecCopy));
-    ASSERT_OK(func->AddKernel({InputType::Array(float64())}, float64(), ExecCopy));
+    ASSERT_OK(func->AddKernel({uint8()}, uint8(), ExecCopyArraySpan));
+    ASSERT_OK(func->AddKernel({int32()}, int32(), ExecCopyArraySpan));
+    ASSERT_OK(func->AddKernel({float64()}, float64(), ExecCopyArraySpan));
     ASSERT_OK(registry->AddFunction(func));
 
     // A version which doesn't want the executor to call PropagateNulls
-    auto func2 = std::make_shared<ScalarFunction>("test_copy_computed_bitmap",
-                                                  Arity::Unary(), /*doc=*/nullptr);
-    ScalarKernel kernel({InputType::Array(uint8())}, uint8(), ExecComputedBitmap);
+    auto func2 = std::make_shared<ScalarFunction>(
+        "test_copy_computed_bitmap", Arity::Unary(), /*doc=*/FunctionDoc::Empty());
+    ScalarKernel kernel({uint8()}, uint8(), ExecComputedBitmap);
     kernel.null_handling = NullHandling::COMPUTED_PREALLOCATE;
     ASSERT_OK(func2->AddKernel(kernel));
     ASSERT_OK(registry->AddFunction(func2));
@@ -705,11 +1032,11 @@ class TestCallScalarFunction : public TestComputeInternals {
     // A function that allocates its own output memory. We have cases for both
     // non-preallocated data and non-preallocated validity bitmap
     auto f1 = std::make_shared<ScalarFunction>("test_nopre_data", Arity::Unary(),
-                                               /*doc=*/nullptr);
-    auto f2 = std::make_shared<ScalarFunction>("test_nopre_validity_or_data",
-                                               Arity::Unary(), /*doc=*/nullptr);
+                                               /*doc=*/FunctionDoc::Empty());
+    auto f2 = std::make_shared<ScalarFunction>(
+        "test_nopre_validity_or_data", Arity::Unary(), /*doc=*/FunctionDoc::Empty());
 
-    ScalarKernel kernel({InputType::Array(uint8())}, uint8(), ExecNoPreallocatedData);
+    ScalarKernel kernel({uint8()}, uint8(), ExecNoPreallocatedData);
     kernel.mem_allocation = MemAllocation::NO_PREALLOCATE;
     ASSERT_OK(f1->AddKernel(kernel));
 
@@ -727,9 +1054,9 @@ class TestCallScalarFunction : public TestComputeInternals {
     // This function's behavior depends on a static parameter that is made
     // available to the kernel's execution function through its Options object
     auto func = std::make_shared<ScalarFunction>("test_stateful", Arity::Unary(),
-                                                 /*doc=*/nullptr);
+                                                 /*doc=*/FunctionDoc::Empty());
 
-    ScalarKernel kernel({InputType::Array(int32())}, int32(), ExecStateful, InitStateful);
+    ScalarKernel kernel({int32()}, int32(), ExecStateful, InitStateful);
     ASSERT_OK(func->AddKernel(kernel));
     ASSERT_OK(registry->AddFunction(func));
   }
@@ -738,9 +1065,8 @@ class TestCallScalarFunction : public TestComputeInternals {
     auto registry = GetFunctionRegistry();
 
     auto func = std::make_shared<ScalarFunction>("test_scalar_add_int32", Arity::Binary(),
-                                                 /*doc=*/nullptr);
-    ASSERT_OK(func->AddKernel({InputType::Scalar(int32()), InputType::Scalar(int32())},
-                              int32(), ExecAddInt32));
+                                                 /*doc=*/FunctionDoc::Empty());
+    ASSERT_OK(func->AddKernel({int32(), int32()}, int32(), ExecAddInt32));
     ASSERT_OK(registry->AddFunction(func));
   }
 };
@@ -760,14 +1086,15 @@ TEST_F(TestCallScalarFunction, ArgumentValidation) {
   ASSERT_RAISES(Invalid, CallFunction("test_copy", args));
 
   // Cannot do scalar
-  args = {Datum(std::make_shared<Int32Scalar>(5))};
-  ASSERT_RAISES(NotImplemented, CallFunction("test_copy", args));
+  Datum d1_scalar(std::make_shared<Int32Scalar>(5));
+  ASSERT_OK_AND_ASSIGN(auto result, CallFunction("test_copy", {d1}));
+  ASSERT_OK_AND_ASSIGN(result, CallFunction("test_copy", {d1_scalar}));
 }
 
 TEST_F(TestCallScalarFunction, PreallocationCases) {
   double null_prob = 0.2;
 
-  auto arr = GetUInt8Array(1000, null_prob);
+  auto arr = GetUInt8Array(100, null_prob);
 
   auto CheckFunction = [&](std::string func_name) {
     ResetContexts();
@@ -792,15 +1119,15 @@ TEST_F(TestCallScalarFunction, PreallocationCases) {
     {
       // Chunksize not multiple of 8
       std::vector<Datum> args = {Datum(arr)};
-      exec_ctx_->set_exec_chunksize(111);
+      exec_ctx_->set_exec_chunksize(11);
       ASSERT_OK_AND_ASSIGN(Datum result, CallFunction(func_name, args, exec_ctx_.get()));
       AssertArraysEqual(*arr, *result.make_array());
     }
 
     // Input is chunked, output has one big chunk
     {
-      auto carr = std::shared_ptr<ChunkedArray>(
-          new ChunkedArray({arr->Slice(0, 100), arr->Slice(100)}));
+      auto carr =
+          std::make_shared<ChunkedArray>(ArrayVector{arr->Slice(0, 10), arr->Slice(10)});
       std::vector<Datum> args = {Datum(carr)};
       ASSERT_OK_AND_ASSIGN(Datum result, CallFunction(func_name, args, exec_ctx_.get()));
       std::shared_ptr<ChunkedArray> actual = result.chunked_array();
@@ -812,14 +1139,14 @@ TEST_F(TestCallScalarFunction, PreallocationCases) {
     {
       std::vector<Datum> args = {Datum(arr)};
       exec_ctx_->set_preallocate_contiguous(false);
-      exec_ctx_->set_exec_chunksize(400);
+      exec_ctx_->set_exec_chunksize(40);
       ASSERT_OK_AND_ASSIGN(Datum result, CallFunction(func_name, args, exec_ctx_.get()));
       ASSERT_EQ(Datum::CHUNKED_ARRAY, result.kind());
       const ChunkedArray& carr = *result.chunked_array();
       ASSERT_EQ(3, carr.num_chunks());
-      AssertArraysEqual(*arr->Slice(0, 400), *carr.chunk(0));
-      AssertArraysEqual(*arr->Slice(400, 400), *carr.chunk(1));
-      AssertArraysEqual(*arr->Slice(800), *carr.chunk(2));
+      AssertArraysEqual(*arr->Slice(0, 40), *carr.chunk(0));
+      AssertArraysEqual(*arr->Slice(40, 40), *carr.chunk(1));
+      AssertArraysEqual(*arr->Slice(80), *carr.chunk(2));
     }
   };
 
