@@ -128,6 +128,12 @@ TEST(TestField, Equals) {
   AssertFieldEqual(f0, f0_with_meta1);
   AssertFieldEqual(f0, f0_with_meta2);
   AssertFieldEqual(f0_with_meta1, f0_with_meta2);
+
+  // operator==(), where check_metadata == false
+  ASSERT_EQ(f0, f0_other);
+  ASSERT_NE(f0, f0_nn);
+  ASSERT_EQ(f0, f0_with_meta1);
+  ASSERT_EQ(f0_with_meta1, f0_with_meta2);
 }
 
 #define ASSERT_COMPATIBLE_IMPL(NAME, TYPE, PLURAL)                        \
@@ -472,9 +478,36 @@ TEST_F(TestSchema, Basics) {
   auto schema3 = std::make_shared<Schema>(fields3);
   AssertSchemaEqual(schema, schema2);
   AssertSchemaNotEqual(schema, schema3);
+  ASSERT_EQ(*schema, *schema2);
+  ASSERT_NE(*schema, *schema3);
 
   ASSERT_EQ(schema->fingerprint(), schema2->fingerprint());
   ASSERT_NE(schema->fingerprint(), schema3->fingerprint());
+
+  auto schema4 = ::arrow::schema({f0}, Endianness::Little);
+  auto schema5 = ::arrow::schema({f0}, Endianness::Little);
+  auto schema6 = ::arrow::schema({f0}, Endianness::Big);
+  auto schema7 = ::arrow::schema({f0});
+
+  AssertSchemaEqual(schema4, schema5);
+  AssertSchemaNotEqual(schema4, schema6);
+#if ARROW_LITTLE_ENDIAN
+  AssertSchemaEqual(schema4, schema7);
+  AssertSchemaNotEqual(schema6, schema7);
+#else
+  AssertSchemaNotEqual(schema4, schema6);
+  AssertSchemaEqual(schema6, schema7);
+#endif
+
+  ASSERT_EQ(schema4->fingerprint(), schema5->fingerprint());
+  ASSERT_NE(schema4->fingerprint(), schema6->fingerprint());
+#if ARROW_LITTLE_ENDIAN
+  ASSERT_EQ(schema4->fingerprint(), schema7->fingerprint());
+  ASSERT_NE(schema6->fingerprint(), schema7->fingerprint());
+#else
+  ASSERT_NE(schema4->fingerprint(), schema7->fingerprint());
+  ASSERT_EQ(schema6->fingerprint(), schema7->fingerprint());
+#endif
 }
 
 TEST_F(TestSchema, ToString) {
@@ -495,14 +528,38 @@ f3: list<item: int16>)";
   ASSERT_EQ(expected, result);
 
   result = schema->ToString(/*print_metadata=*/true);
+  std::string expected_with_metadata = expected + R"(
+-- metadata --
+foo: bar)";
+
+  ASSERT_EQ(expected_with_metadata, result);
+
+  // With swapped endianness
+#if ARROW_LITTLE_ENDIAN
+  schema = schema->WithEndianness(Endianness::Big);
   expected = R"(f0: int32
 f1: uint8 not null
 f2: string
 f3: list<item: int16>
+-- endianness: big --)";
+#else
+  schema = schema->WithEndianness(Endianness::Little);
+  expected = R"(f0: int32
+f1: uint8 not null
+f2: string
+f3: list<item: int16>
+-- endianness: little --)";
+#endif
+
+  result = schema->ToString();
+  ASSERT_EQ(expected, result);
+
+  result = schema->ToString(/*print_metadata=*/true);
+  expected_with_metadata = expected + R"(
 -- metadata --
 foo: bar)";
 
-  ASSERT_EQ(expected, result);
+  ASSERT_EQ(expected_with_metadata, result);
 }
 
 TEST_F(TestSchema, GetFieldByName) {
@@ -1298,6 +1355,23 @@ TEST(TestDayTimeIntervalType, ToString) {
   ASSERT_EQ("day_time_interval", t1->ToString());
 }
 
+TEST(TestMonthDayNanoIntervalType, Equals) {
+  MonthDayNanoIntervalType t1;
+  MonthDayNanoIntervalType t2;
+  MonthIntervalType t3;
+  DayTimeIntervalType t4;
+
+  AssertTypeEqual(t1, t2);
+  AssertTypeNotEqual(t1, t3);
+  AssertTypeNotEqual(t1, t4);
+}
+
+TEST(TestMonthDayNanoIntervalType, ToString) {
+  auto t1 = month_day_nano_interval();
+
+  ASSERT_EQ("month_day_nano_interval", t1->ToString());
+}
+
 TEST(TestDurationType, Equals) {
   DurationType t1;
   DurationType t2;
@@ -1533,6 +1607,38 @@ TEST(TestStructType, TestFieldsDifferOnlyInMetadata) {
   ASSERT_NE(s0.metadata_fingerprint(), s1.metadata_fingerprint());
 }
 
+TEST(TestStructType, FieldModifierMethods) {
+  auto f0 = field("f0", int32());
+  auto f1 = field("f1", utf8());
+
+  std::vector<std::shared_ptr<Field>> fields = {f0, f1};
+
+  StructType struct_type(fields);
+
+  ASSERT_OK_AND_ASSIGN(auto new_struct, struct_type.AddField(1, field("f2", int8())));
+  ASSERT_EQ(3, new_struct->num_fields());
+  ASSERT_EQ(1, new_struct->GetFieldIndex("f2"));
+
+  ASSERT_OK_AND_ASSIGN(new_struct, new_struct->RemoveField(1));
+  ASSERT_EQ(2, new_struct->num_fields());
+  ASSERT_EQ(-1, new_struct->GetFieldIndex("f2"));  // No f2 after removal
+
+  ASSERT_OK_AND_ASSIGN(new_struct, new_struct->SetField(1, field("f2", int8())));
+  ASSERT_EQ(2, new_struct->num_fields());
+  ASSERT_EQ(1, new_struct->GetFieldIndex("f2"));
+  ASSERT_EQ(int8(), new_struct->GetFieldByName("f2")->type());
+
+  EXPECT_RAISES_WITH_MESSAGE_THAT(Invalid,
+                                  testing::HasSubstr("Invalid column index to add field"),
+                                  new_struct->AddField(5, field("f5", int8())));
+  EXPECT_RAISES_WITH_MESSAGE_THAT(
+      Invalid, testing::HasSubstr("Invalid column index to remove field"),
+      new_struct->RemoveField(-1));
+  EXPECT_RAISES_WITH_MESSAGE_THAT(Invalid,
+                                  testing::HasSubstr("Invalid column index to set field"),
+                                  new_struct->SetField(2, field("f5", int8())));
+}
+
 TEST(TestUnionType, Basics) {
   auto f0_type = int32();
   auto f0 = field("f0", f0_type);
@@ -1613,160 +1719,6 @@ TEST(TestDictionaryType, Equals) {
   AssertTypeNotEqual(*t5, *t6);
 }
 
-void CheckTransposeMap(const Buffer& map, std::vector<int32_t> expected) {
-  AssertBufferEqual(map, *Buffer::Wrap(expected));
-}
-
-TEST(TestDictionaryType, UnifyNumeric) {
-  auto dict_ty = int64();
-
-  auto d1 = ArrayFromJSON(dict_ty, "[3, 4, 7]");
-  auto d2 = ArrayFromJSON(dict_ty, "[1, 7, 4, 8]");
-  auto d3 = ArrayFromJSON(dict_ty, "[1, -200]");
-
-  auto expected = dictionary(int8(), dict_ty);
-  auto expected_dict = ArrayFromJSON(dict_ty, "[3, 4, 7, 1, 8, -200]");
-
-  ASSERT_OK_AND_ASSIGN(auto unifier, DictionaryUnifier::Make(dict_ty));
-
-  std::shared_ptr<DataType> out_type;
-  std::shared_ptr<Array> out_dict;
-
-  ASSERT_OK(unifier->Unify(*d1));
-  ASSERT_OK(unifier->Unify(*d2));
-  ASSERT_OK(unifier->Unify(*d3));
-
-  ASSERT_RAISES(Invalid, unifier->Unify(*ArrayFromJSON(int32(), "[1, -200]")));
-
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  std::shared_ptr<Buffer> b1, b2, b3;
-
-  ASSERT_OK(unifier->Unify(*d1, &b1));
-  ASSERT_OK(unifier->Unify(*d2, &b2));
-  ASSERT_OK(unifier->Unify(*d3, &b3));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  CheckTransposeMap(*b1, {0, 1, 2});
-  CheckTransposeMap(*b2, {3, 2, 1, 4});
-  CheckTransposeMap(*b3, {3, 5});
-}
-
-TEST(TestDictionaryType, UnifyString) {
-  auto dict_ty = utf8();
-
-  auto t1 = dictionary(int16(), dict_ty);
-  auto d1 = ArrayFromJSON(dict_ty, "[\"foo\", \"bar\"]");
-
-  auto t2 = dictionary(int32(), dict_ty);
-  auto d2 = ArrayFromJSON(dict_ty, "[\"quux\", \"foo\"]");
-
-  auto expected = dictionary(int8(), dict_ty);
-  auto expected_dict = ArrayFromJSON(dict_ty, "[\"foo\", \"bar\", \"quux\"]");
-
-  ASSERT_OK_AND_ASSIGN(auto unifier, DictionaryUnifier::Make(dict_ty));
-
-  std::shared_ptr<DataType> out_type;
-  std::shared_ptr<Array> out_dict;
-  ASSERT_OK(unifier->Unify(*d1));
-  ASSERT_OK(unifier->Unify(*d2));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  std::shared_ptr<Buffer> b1, b2;
-
-  ASSERT_OK(unifier->Unify(*d1, &b1));
-  ASSERT_OK(unifier->Unify(*d2, &b2));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  CheckTransposeMap(*b1, {0, 1});
-  CheckTransposeMap(*b2, {2, 0});
-}
-
-TEST(TestDictionaryType, UnifyFixedSizeBinary) {
-  auto type = fixed_size_binary(3);
-
-  std::string data = "foobarbazqux";
-  auto buf = std::make_shared<Buffer>(data);
-  // ["foo", "bar"]
-  auto dict1 = std::make_shared<FixedSizeBinaryArray>(type, 2, SliceBuffer(buf, 0, 6));
-  auto t1 = dictionary(int16(), type);
-  // ["bar", "baz", "qux"]
-  auto dict2 = std::make_shared<FixedSizeBinaryArray>(type, 3, SliceBuffer(buf, 3, 9));
-  auto t2 = dictionary(int16(), type);
-
-  // ["foo", "bar", "baz", "qux"]
-  auto expected_dict = std::make_shared<FixedSizeBinaryArray>(type, 4, buf);
-  auto expected = dictionary(int8(), type);
-
-  ASSERT_OK_AND_ASSIGN(auto unifier, DictionaryUnifier::Make(type));
-  std::shared_ptr<DataType> out_type;
-  std::shared_ptr<Array> out_dict;
-  ASSERT_OK(unifier->Unify(*dict1));
-  ASSERT_OK(unifier->Unify(*dict2));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  std::shared_ptr<Buffer> b1, b2;
-  ASSERT_OK(unifier->Unify(*dict1, &b1));
-  ASSERT_OK(unifier->Unify(*dict2, &b2));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-
-  CheckTransposeMap(*b1, {0, 1});
-  CheckTransposeMap(*b2, {1, 2, 3});
-}
-
-TEST(TestDictionaryType, UnifyLarge) {
-  // Unifying "large" dictionary types should choose the right index type
-  std::shared_ptr<Array> dict1, dict2, expected_dict;
-
-  Int32Builder builder;
-  ASSERT_OK(builder.Reserve(120));
-  for (int32_t i = 0; i < 120; ++i) {
-    builder.UnsafeAppend(i);
-  }
-  ASSERT_OK(builder.Finish(&dict1));
-  ASSERT_EQ(dict1->length(), 120);
-  auto t1 = dictionary(int8(), int32());
-
-  ASSERT_OK(builder.Reserve(30));
-  for (int32_t i = 110; i < 140; ++i) {
-    builder.UnsafeAppend(i);
-  }
-  ASSERT_OK(builder.Finish(&dict2));
-  ASSERT_EQ(dict2->length(), 30);
-  auto t2 = dictionary(int8(), int32());
-
-  ASSERT_OK(builder.Reserve(140));
-  for (int32_t i = 0; i < 140; ++i) {
-    builder.UnsafeAppend(i);
-  }
-  ASSERT_OK(builder.Finish(&expected_dict));
-  ASSERT_EQ(expected_dict->length(), 140);
-
-  // int8 would be too narrow to hold all possible index values
-  auto expected = dictionary(int16(), int32());
-
-  ASSERT_OK_AND_ASSIGN(auto unifier, DictionaryUnifier::Make(int32()));
-  std::shared_ptr<DataType> out_type;
-  std::shared_ptr<Array> out_dict;
-  ASSERT_OK(unifier->Unify(*dict1));
-  ASSERT_OK(unifier->Unify(*dict2));
-  ASSERT_OK(unifier->GetResult(&out_type, &out_dict));
-  ASSERT_TRUE(out_type->Equals(*expected));
-  ASSERT_TRUE(out_dict->Equals(*expected_dict));
-}
-
 TEST(TypesTest, TestDecimal128Small) {
   Decimal128Type t1(8, 4);
 
@@ -1774,7 +1726,7 @@ TEST(TypesTest, TestDecimal128Small) {
   EXPECT_EQ(t1.precision(), 8);
   EXPECT_EQ(t1.scale(), 4);
 
-  EXPECT_EQ(t1.ToString(), std::string("decimal(8, 4)"));
+  EXPECT_EQ(t1.ToString(), std::string("decimal128(8, 4)"));
 
   // Test properties
   EXPECT_EQ(t1.byte_width(), 16);
@@ -1788,7 +1740,7 @@ TEST(TypesTest, TestDecimal128Medium) {
   EXPECT_EQ(t1.precision(), 12);
   EXPECT_EQ(t1.scale(), 5);
 
-  EXPECT_EQ(t1.ToString(), std::string("decimal(12, 5)"));
+  EXPECT_EQ(t1.ToString(), std::string("decimal128(12, 5)"));
 
   // Test properties
   EXPECT_EQ(t1.byte_width(), 16);
@@ -1802,7 +1754,7 @@ TEST(TypesTest, TestDecimal128Large) {
   EXPECT_EQ(t1.precision(), 27);
   EXPECT_EQ(t1.scale(), 7);
 
-  EXPECT_EQ(t1.ToString(), std::string("decimal(27, 7)"));
+  EXPECT_EQ(t1.ToString(), std::string("decimal128(27, 7)"));
 
   // Test properties
   EXPECT_EQ(t1.byte_width(), 16);
@@ -1877,4 +1829,47 @@ TEST(TypesTest, TestDecimalEquals) {
   AssertTypeNotEqual(t5, t10);
 }
 
+#define TEST_PREDICATE(all_types, type_predicate)                 \
+  for (auto type : all_types) {                                   \
+    ASSERT_EQ(type_predicate(type->id()), type_predicate(*type)); \
+  }
+
+TEST(TypesTest, TestMembership) {
+  std::vector<std::shared_ptr<DataType>> all_types;
+  for (auto type : NumericTypes()) {
+    all_types.push_back(type);
+  }
+  for (auto type : TemporalTypes()) {
+    all_types.push_back(type);
+  }
+  for (auto type : IntervalTypes()) {
+    all_types.push_back(type);
+  }
+  for (auto type : PrimitiveTypes()) {
+    all_types.push_back(type);
+  }
+  TEST_PREDICATE(all_types, is_integer);
+  TEST_PREDICATE(all_types, is_signed_integer);
+  TEST_PREDICATE(all_types, is_unsigned_integer);
+  TEST_PREDICATE(all_types, is_floating);
+  TEST_PREDICATE(all_types, is_numeric);
+  TEST_PREDICATE(all_types, is_decimal);
+  TEST_PREDICATE(all_types, is_primitive);
+  TEST_PREDICATE(all_types, is_base_binary_like);
+  TEST_PREDICATE(all_types, is_binary_like);
+  TEST_PREDICATE(all_types, is_large_binary_like);
+  TEST_PREDICATE(all_types, is_binary);
+  TEST_PREDICATE(all_types, is_string);
+  TEST_PREDICATE(all_types, is_temporal);
+  TEST_PREDICATE(all_types, is_interval);
+  TEST_PREDICATE(all_types, is_dictionary);
+  TEST_PREDICATE(all_types, is_fixed_size_binary);
+  TEST_PREDICATE(all_types, is_fixed_width);
+  TEST_PREDICATE(all_types, is_list_like);
+  TEST_PREDICATE(all_types, is_nested);
+  TEST_PREDICATE(all_types, is_union);
+}
+
+#undef TEST_PREDICATE
+
 }  // namespace arrow
