@@ -1,5 +1,4 @@
-#!/usr/bin/env python
-
+#!/usr/bin/env python3
 #
 # Licensed to the Apache Software Foundation (ASF) under one or more
 # contributor license agreements.  See the NOTICE file distributed with
@@ -19,19 +18,23 @@
 
 # Utility for creating well-formed pull request merges and pushing them to
 # Apache.
-#   usage: ./merge_arrow_pr.py    (see config env vars below)
+#   usage: ./merge_arrow_pr.py  <pr-number>  (see config env vars below)
 #
-# This utility assumes you already have a local Arrow git clone and that you
-# have added remotes corresponding to both (i) the GitHub Apache Arrow mirror
-# and (ii) the apache git repo.
+# This utility assumes:
+#   - you already have a local Arrow git clone
+#   - you have added remotes corresponding to both:
+#       (i) the GitHub Apache Arrow mirror
+#       (ii) the Apache git repo
 #
 # There are several pieces of authorization possibly needed via environment
-# variables
+# variables.
 #
-# APACHE_JIRA_USERNAME: your Apache JIRA id
-# APACHE_JIRA_PASSWORD: your Apache JIRA password
-# ARROW_GITHUB_API_TOKEN: a GitHub API token to use for API requests (to avoid
-# rate limiting)
+# Configuration environment variables:
+#   - APACHE_JIRA_TOKEN: your Apache JIRA Personal Access Token
+#   - ARROW_GITHUB_API_TOKEN: a GitHub API token to use for API requests
+#   - PR_REMOTE_NAME: the name of the remote to the Apache git repo (set to
+#                     'apache' by default)
+#   - DEBUG: use for testing to avoid pushing to apache (0 by default)
 
 import configparser
 import os
@@ -50,7 +53,7 @@ try:
     import jira.exceptions
 except ImportError:
     print("Could not find jira library. "
-          "Run 'sudo pip install jira' to install.")
+          "Run 'pip install jira' to install.")
     print("Exiting without trying to close the associated JIRA.")
     sys.exit(1)
 
@@ -65,14 +68,12 @@ if DEBUG:
     print("**************** DEBUGGING ****************")
 
 
-# Prefix added to temporary branches
-BRANCH_PREFIX = "PR_TOOL"
 JIRA_API_BASE = "https://issues.apache.org/jira"
 
 
 def get_json(url, headers=None):
-    req = requests.get(url, headers=headers)
-    return req.json()
+    response = requests.get(url, headers=headers)
+    return response.json()
 
 
 def run_cmd(cmd):
@@ -95,21 +96,6 @@ def run_cmd(cmd):
     return output
 
 
-original_head = run_cmd("git rev-parse HEAD")[:8]
-
-
-def clean_up():
-    print("Restoring head pointer to %s" % original_head)
-    run_cmd("git checkout %s" % original_head)
-
-    branches = run_cmd("git branch").replace(" ", "").split("\n")
-
-    for branch in [x for x in branches
-                   if x.startswith(BRANCH_PREFIX)]:
-        print("Deleting local branch %s" % branch)
-        run_cmd("git branch -D %s" % branch)
-
-
 _REGEX_CI_DIRECTIVE = re.compile(r'\[[^\]]*\]')
 
 
@@ -129,7 +115,7 @@ def fix_version_from_branch(branch, versions):
         return [x for x in versions if x.name.startswith(branch_ver)][-1]
 
 
-# We can merge both ARROW and PARQUET patchesa
+# We can merge both ARROW and PARQUET patches
 SUPPORTED_PROJECTS = ['ARROW', 'PARQUET']
 PR_TITLE_REGEXEN = [(project, re.compile(r'^(' + project + r'-[0-9]+)\b.*$'))
                     for project in SUPPORTED_PROJECTS]
@@ -152,16 +138,23 @@ class JiraIssue(object):
     def current_fix_versions(self):
         return self.issue.fields.fixVersions
 
-    def get_candidate_fix_versions(self, merge_branches=('master',)):
+    @classmethod
+    def sort_versions(cls, versions):
+        def version_tuple(x):
+            # Parquet versions are something like cpp-1.2.0
+            numeric_version = x.name.split("-", 1)[-1]
+            return tuple(int(_) for _ in numeric_version.split("."))
+        return sorted(versions, key=version_tuple, reverse=True)
+
+    def get_candidate_fix_versions(self, merge_branches=('master',),
+                                   maintenance_branches=()):
         # Only suggest versions starting with a number, like 0.x but not JS-0.x
         all_versions = self.jira_con.project_versions(self.project)
         unreleased_versions = [x for x in all_versions
                                if not x.raw['released']]
 
-        unreleased_versions = sorted(unreleased_versions,
-                                     key=lambda x: x.name, reverse=True)
-
         mainline_versions = self._filter_mainline_versions(unreleased_versions)
+        mainline_versions = self.sort_versions(mainline_versions)
 
         mainline_non_patch_versions = []
         for v in mainline_versions:
@@ -173,12 +166,19 @@ class JiraIssue(object):
             # If there is a non-patch release, suggest that instead
             mainline_versions = mainline_non_patch_versions
 
+        mainline_versions = self._filter_maintenance_versions(
+            mainline_versions, maintenance_branches
+        )
         default_fix_versions = [
             fix_version_from_branch(x, mainline_versions).name
             for x in merge_branches]
 
         return all_versions, default_fix_versions
 
+    def _filter_maintenance_versions(self, versions, maintenance_branches):
+        return [v for v in versions
+                if f"maint-{v.name}" not in maintenance_branches]
+
     def _filter_mainline_versions(self, versions):
         if self.project == 'PARQUET':
             mainline_regex = re.compile(r'cpp-\d.*')
@@ -249,20 +249,52 @@ URL\t\t{}/{}""".format(jira_id, summary, assignee, components, status,
 
 class GitHubAPI(object):
 
-    def __init__(self, project_name):
+    def __init__(self, project_name, cmd):
         self.github_api = ("https://api.github.com/repos/apache/{0}"
                            .format(project_name))
 
-        token = os.environ.get('ARROW_GITHUB_API_TOKEN', None)
-        if token:
-            self.headers = {'Authorization': 'token {0}'.format(token)}
-        else:
-            self.headers = None
+        token = None
+        config = load_configuration()
+        if "github" in config.sections():
+            token = config["github"]["api_token"]
+        if not token:
+            token = os.environ.get('ARROW_GITHUB_API_TOKEN')
+        if not token:
+            token = cmd.prompt('Env ARROW_GITHUB_API_TOKEN not set, '
+                               'please enter your GitHub API token '
+                               '(GitHub personal access token):')
+        headers = {
+            'Accept': 'application/vnd.github.v3+json',
+            'Authorization': 'token {0}'.format(token),
+        }
+        self.headers = headers
 
     def get_pr_data(self, number):
         return get_json("%s/pulls/%s" % (self.github_api, number),
                         headers=self.headers)
 
+    def get_pr_commits(self, number):
+        return get_json("%s/pulls/%s/commits" % (self.github_api, number),
+                        headers=self.headers)
+
+    def get_branches(self):
+        return get_json("%s/branches" % (self.github_api),
+                        headers=self.headers)
+
+    def merge_pr(self, number, commit_title, commit_message):
+        url = f'{self.github_api}/pulls/{number}/merge'
+        payload = {
+            'commit_title': commit_title,
+            'commit_message': commit_message,
+            'merge_method': 'squash',
+        }
+        response = requests.put(url, headers=self.headers, json=payload)
+        result = response.json()
+        if response.status_code != 200 and 'merged' not in result:
+            result['merged'] = False
+            result['message'] += f': {url}'
+        return result
+
 
 class CommandInput(object):
     """
@@ -270,7 +302,6 @@ class CommandInput(object):
     """
 
     def fail(self, msg):
-        clean_up()
         raise Exception(msg)
 
     def prompt(self, prompt):
@@ -294,6 +325,7 @@ class PullRequest(object):
 
     def __init__(self, cmd, github_api, git_remote, jira_con, number):
         self.cmd = cmd
+        self._github_api = github_api
         self.git_remote = git_remote
         self.con = jira_con
         self.number = number
@@ -329,7 +361,15 @@ class PullRequest(object):
     def is_mergeable(self):
         return bool(self._pr_data["mergeable"])
 
+    @property
+    def maintenance_branches(self):
+        return [x["name"] for x in self._github_api.get_branches()
+                if x["name"].startswith("maint-")]
+
     def _get_jira(self):
+        if self.title.startswith("MINOR:"):
+            return None
+
         jira_id = None
         for project, regex in PR_TITLE_REGEXEN:
             m = regex.search(self.title)
@@ -337,7 +377,7 @@ class PullRequest(object):
                 jira_id = m.group(1)
                 break
 
-        if jira_id is None and not self.title.startswith("MINOR:"):
+        if jira_id is None:
             options = ' or '.join('{0}-XXX'.format(project)
                                   for project in SUPPORTED_PROJECTS)
             self.cmd.fail("PR title should be prefixed by a jira id "
@@ -349,32 +389,26 @@ class PullRequest(object):
         """
         merge the requested PR and return the merge hash
         """
-        pr_branch_name = "%s_MERGE_PR_%s" % (BRANCH_PREFIX, self.number)
-        target_branch_name = "%s_MERGE_PR_%s_%s" % (BRANCH_PREFIX,
-                                                    self.number,
-                                                    self.target_ref.upper())
-        run_cmd("git fetch %s pull/%s/head:%s" % (self.git_remote,
-                                                  self.number,
-                                                  pr_branch_name))
-        run_cmd("git fetch %s %s:%s" % (self.git_remote, self.target_ref,
-                                        target_branch_name))
-        run_cmd("git checkout %s" % target_branch_name)
+        commits = self._github_api.get_pr_commits(self.number)
 
-        had_conflicts = False
-        try:
-            run_cmd(['git', 'merge', pr_branch_name, '--ff', '--squash'])
-        except Exception as e:
-            msg = ("Error merging: %s\nWould you like to "
-                   "manually fix-up this merge?" % e)
-            self.cmd.continue_maybe(msg)
-            msg = ("Okay, please fix any conflicts and 'git add' "
-                   "conflicting files... Finished?")
-            self.cmd.continue_maybe(msg)
-            had_conflicts = True
+        def format_commit_author(commit):
+            author = commit['commit']['author']
+            name = author['name']
+            email = author['email']
+            return f'{name} <{email}>'
+        commit_authors = [format_commit_author(commit) for commit in commits]
+        co_authored_by_re = re.compile(
+            r'^Co-authored-by:\s*(.*)', re.MULTILINE)
 
-        commit_authors = run_cmd(['git', 'log', 'HEAD..%s' % pr_branch_name,
-                                 '--pretty=format:%an <%ae>']).split("\n")
-        distinct_authors = sorted(set(commit_authors),
+        def extract_co_authors(commit):
+            message = commit['commit']['message']
+            return co_authored_by_re.findall(message)
+        commit_co_authors = []
+        for commit in commits:
+            commit_co_authors.extend(extract_co_authors(commit))
+
+        all_commit_authors = commit_authors + commit_co_authors
+        distinct_authors = sorted(set(all_commit_authors),
                                   key=lambda x: commit_authors.count(x),
                                   reverse=True)
 
@@ -382,74 +416,51 @@ class PullRequest(object):
             print("Author {}: {}".format(i + 1, author))
 
         if len(distinct_authors) > 1:
-            primary_author, distinct_authors = get_primary_author(
+            primary_author, distinct_other_authors = get_primary_author(
                 self.cmd, distinct_authors)
         else:
             # If there is only one author, do not prompt for a lead author
-            primary_author = distinct_authors[0]
+            primary_author = distinct_authors.pop()
+            distinct_other_authors = []
 
-        merge_message_flags = []
-
-        merge_message_flags += ["-m", self.title]
+        commit_title = f'{self.title} (#{self.number})'
+        commit_message_chunks = []
         if self.body is not None:
-            merge_message_flags += ["-m", self.body]
+            commit_message_chunks.append(self.body)
 
         committer_name = run_cmd("git config --get user.name").strip()
         committer_email = run_cmd("git config --get user.email").strip()
 
-        authors = ("Authored-by:" if len(distinct_authors) == 1
+        authors = ("Authored-by:" if len(distinct_other_authors) == 0
                    else "Lead-authored-by:")
-        authors += " %s" % (distinct_authors.pop(0))
+        authors += " %s" % primary_author
         if len(distinct_authors) > 0:
             authors += "\n" + "\n".join(["Co-authored-by: %s" % a
-                                         for a in distinct_authors])
+                                         for a in distinct_other_authors])
         authors += "\n" + "Signed-off-by: %s <%s>" % (committer_name,
                                                       committer_email)
+        commit_message_chunks.append(authors)
 
-        if had_conflicts:
-            committer_name = run_cmd("git config --get user.name").strip()
-            committer_email = run_cmd("git config --get user.email").strip()
-            message = ("This patch had conflicts when merged, "
-                       "resolved by\nCommitter: %s <%s>" %
-                       (committer_name, committer_email))
-            merge_message_flags += ["-m", message]
-
-        # The string "Closes #%s" string is required for GitHub to correctly
-        # close the PR
-        merge_message_flags += [
-            "-m",
-            "Closes #%s from %s"
-            % (self.number, self.description)]
-        merge_message_flags += ["-m", authors]
+        commit_message = "\n\n".join(commit_message_chunks)
 
         if DEBUG:
-            print("\n".join(merge_message_flags))
+            print(commit_title)
+            print()
+            print(commit_message)
 
-        run_cmd(['git', 'commit',
-                 '--no-verify',  # do not run commit hooks
-                 '--author="%s"' % primary_author] +
-                merge_message_flags)
+        if DEBUG:
+            merge_hash = None
+        else:
+            result = self._github_api.merge_pr(self.number,
+                                               commit_title,
+                                               commit_message)
+            if not result['merged']:
+                message = result['message']
+                self.cmd.fail(f'Failed to merge pull request: {message}')
+            merge_hash = result['sha']
 
-        self.cmd.continue_maybe("Merge complete (local ref %s). Push to %s?"
-                                % (target_branch_name, self.git_remote))
-
-        try:
-            push_cmd = ('git push %s %s:%s' % (self.git_remote,
-                                               target_branch_name,
-                                               self.target_ref))
-            if DEBUG:
-                print(push_cmd)
-            else:
-                run_cmd(push_cmd)
-        except Exception as e:
-            clean_up()
-            self.cmd.fail("Exception while pushing: %s" % e)
-
-        merge_hash = run_cmd("git rev-parse %s" % target_branch_name)[:8]
-        clean_up()
         print("Pull request #%s merged!" % self.number)
         print("Merge hash: %s" % merge_hash)
-        return merge_hash
 
 
 def get_primary_author(cmd, distinct_authors):
@@ -461,7 +472,7 @@ def get_primary_author(cmd, distinct_authors):
             "\"name <email>\" [%s]: " % distinct_authors[0])
 
         if primary_author == "":
-            return distinct_authors[0], distinct_authors
+            return distinct_authors[0], distinct_authors[1:]
 
         if author_pat.match(primary_author):
             break
@@ -469,15 +480,16 @@ def get_primary_author(cmd, distinct_authors):
 
     # When primary author is specified manually, de-dup it from
     # author list and put it at the head of author list.
-    distinct_authors = [x for x in distinct_authors
-                        if x != primary_author]
-    distinct_authors = [primary_author] + distinct_authors
-    return primary_author, distinct_authors
+    distinct_other_authors = [x for x in distinct_authors
+                              if x != primary_author]
+    return primary_author, distinct_other_authors
 
 
-def prompt_for_fix_version(cmd, jira_issue):
+def prompt_for_fix_version(cmd, jira_issue, maintenance_branches=()):
     (all_versions,
-     default_fix_versions) = jira_issue.get_candidate_fix_versions()
+     default_fix_versions) = jira_issue.get_candidate_fix_versions(
+        maintenance_branches=maintenance_branches
+    )
 
     default_fix_versions = ",".join(default_fix_versions)
 
@@ -504,47 +516,28 @@ def load_configuration():
 
 
 def get_credentials(cmd):
-    username, password = None, None
+    token = None
 
     config = load_configuration()
     if "jira" in config.sections():
-        username = config["jira"].get("username")
-        password = config["jira"].get("password")
+        token = config["jira"].get("token")
 
     # Fallback to environment variables
-    if not username:
-        username = os.environ.get("APACHE_JIRA_USERNAME")
-
-    if not password:
-        password = os.environ.get("APACHE_JIRA_PASSWORD")
+    if not token:
+        token = os.environ.get("APACHE_JIRA_TOKEN")
 
     # Fallback to user tty prompt
-    if not username:
-        username = cmd.prompt("Env APACHE_JIRA_USERNAME not set, "
-                              "please enter your JIRA username:")
+    if not token:
+        token = cmd.prompt("Env APACHE_JIRA_TOKEN not set, "
+                           "please enter your Jira API token "
+                           "(Jira personal access token):")
 
-    if not password:
-        password = cmd.getpass("Env APACHE_JIRA_PASSWORD not set, "
-                               "please enter your JIRA password:")
-
-    return (username, password)
+    return token
 
 
 def connect_jira(cmd):
-    try:
-        return jira.client.JIRA(options={'server': JIRA_API_BASE},
-                                basic_auth=get_credentials(cmd))
-    except jira.exceptions.JIRAError as e:
-        if "CAPTCHA_CHALLENGE" in e.text:
-            print("")
-            print("It looks like you need to answer a captcha challenge for "
-                  "this account (probably due to a login attempt with an "
-                  "incorrect password). Please log in at "
-                  "https://issues.apache.org/jira and complete the captcha "
-                  "before running this tool again.")
-            print("Exiting.")
-            sys.exit(1)
-        raise e
+    return jira.client.JIRA(options={'server': JIRA_API_BASE},
+                            token_auth=get_credentials(cmd))
 
 
 def get_pr_num():
@@ -567,25 +560,23 @@ def cli():
 
     os.chdir(ARROW_HOME)
 
-    github_api = GitHubAPI(PROJECT_NAME)
+    github_api = GitHubAPI(PROJECT_NAME, cmd)
 
     jira_con = connect_jira(cmd)
     pr = PullRequest(cmd, github_api, PR_REMOTE_NAME, jira_con, pr_num)
 
     if pr.is_merged:
-        print("Pull request %s has already been merged")
+        print("Pull request %s has already been merged" % pr_num)
         sys.exit(0)
 
     if not pr.is_mergeable:
-        msg = ("Pull request %s is not mergeable in its current form.\n"
-               % pr_num + "Continue? (experts only!)")
-        cmd.continue_maybe(msg)
+        print("Pull request %s is not mergeable in its current form" % pr_num)
+        sys.exit(1)
 
     pr.show()
 
     cmd.continue_maybe("Proceed with merging pull request #%s?" % pr_num)
 
-    # merged hash not used
     pr.merge()
 
     if pr.jira_issue is None:
@@ -599,7 +590,8 @@ def cli():
            "https://github.com/apache/" + PROJECT_NAME + "/pull",
            pr_num))
 
-    fix_versions_json = prompt_for_fix_version(cmd, pr.jira_issue)
+    fix_versions_json = prompt_for_fix_version(cmd, pr.jira_issue,
+                                               pr.maintenance_branches)
     pr.jira_issue.resolve(fix_versions_json, jira_comment)
 
 
