@@ -34,7 +34,7 @@ else:
     from distutils import sysconfig
 
 import pkg_resources
-from setuptools import setup, Extension, Distribution
+from setuptools import setup, Extension, Distribution, find_namespace_packages
 
 from Cython.Distutils import build_ext as _build_ext
 import Cython
@@ -42,8 +42,8 @@ import Cython
 # Check if we're running 64-bit Python
 is_64_bit = sys.maxsize > 2**32
 
-if Cython.__version__ < '0.29':
-    raise Exception('Please upgrade to Cython 0.29 or newer')
+if Cython.__version__ < '0.29.22':
+    raise Exception('Please upgrade to Cython 0.29.22 or newer')
 
 setup_dir = os.path.abspath(os.path.dirname(__file__))
 
@@ -93,7 +93,9 @@ class build_ext(_build_ext):
         _build_ext.build_extensions(self)
 
     def run(self):
-        self._run_cmake()
+        pyarrow_cpp_home = pjoin(os.getcwd(), 'build', 'dist')
+        self._run_cmake_pyarrow_cpp(pyarrow_cpp_home)
+        self._run_cmake(pyarrow_cpp_home)
         _build_ext.run(self)
 
     # adapted from cmake_build_ext in dynd-python
@@ -108,8 +110,13 @@ class build_ext(_build_ext):
                       'namespace of boost (default: boost)'),
                      ('with-cuda', None, 'build the Cuda extension'),
                      ('with-flight', None, 'build the Flight extension'),
+                     ('with-substrait', None, 'build the Substrait extension'),
                      ('with-dataset', None, 'build the Dataset extension'),
                      ('with-parquet', None, 'build the Parquet extension'),
+                     ('with-parquet-encryption', None,
+                      'build the Parquet encryption extension'),
+                     ('with-gcs', None,
+                      'build the Google Cloud Storage (GCS) extension'),
                      ('with-s3', None, 'build the Amazon S3 extension'),
                      ('with-static-parquet', None, 'link parquet statically'),
                      ('with-static-boost', None, 'link boost statically'),
@@ -152,12 +159,16 @@ class build_ext(_build_ext):
             if not hasattr(sys, 'gettotalrefcount'):
                 self.build_type = 'release'
 
+        self.with_gcs = strtobool(
+            os.environ.get('PYARROW_WITH_GCS', '0'))
         self.with_s3 = strtobool(
             os.environ.get('PYARROW_WITH_S3', '0'))
         self.with_hdfs = strtobool(
             os.environ.get('PYARROW_WITH_HDFS', '0'))
         self.with_cuda = strtobool(
             os.environ.get('PYARROW_WITH_CUDA', '0'))
+        self.with_substrait = strtobool(
+            os.environ.get('PYARROW_WITH_SUBSTRAIT', '0'))
         self.with_flight = strtobool(
             os.environ.get('PYARROW_WITH_FLIGHT', '0'))
         self.with_dataset = strtobool(
@@ -166,6 +177,8 @@ class build_ext(_build_ext):
             os.environ.get('PYARROW_WITH_PARQUET', '0'))
         self.with_static_parquet = strtobool(
             os.environ.get('PYARROW_WITH_STATIC_PARQUET', '0'))
+        self.with_parquet_encryption = strtobool(
+            os.environ.get('PYARROW_WITH_PARQUET_ENCRYPTION', '0'))
         self.with_static_boost = strtobool(
             os.environ.get('PYARROW_WITH_STATIC_BOOST', '0'))
         self.with_plasma = strtobool(
@@ -189,6 +202,9 @@ class build_ext(_build_ext):
         self.bundle_plasma_executable = strtobool(
             os.environ.get('PYARROW_BUNDLE_PLASMA_EXECUTABLE', '1'))
 
+        self.with_parquet_encryption = (self.with_parquet_encryption and
+                                        self.with_parquet)
+
     CYTHON_MODULE_NAMES = [
         'lib',
         '_fs',
@@ -199,33 +215,113 @@ class build_ext(_build_ext):
         '_flight',
         '_dataset',
         '_dataset_orc',
+        '_dataset_parquet',
+        '_exec_plan',
         '_feather',
         '_parquet',
+        '_parquet_encryption',
+        '_pyarrow_cpp_tests',
         '_orc',
         '_plasma',
+        '_gcsfs',
         '_s3fs',
+        '_substrait',
         '_hdfs',
         '_hdfsio',
         'gandiva']
 
-    def _run_cmake(self):
+    def _run_cmake_pyarrow_cpp(self, pyarrow_cpp_home):
         # check if build_type is correctly passed / set
-        if self.build_type.lower() not in ('release', 'debug'):
+        if self.build_type.lower() not in ('release', 'debug',
+                                           'relwithdebinfo'):
             raise ValueError("--build-type (or PYARROW_BUILD_TYPE) needs to "
-                             "be 'release' or 'debug'")
+                             "be 'release', 'debug' or 'relwithdebinfo'")
+
+        # The directory containing this setup.py
+        source = os.path.dirname(os.path.abspath(__file__))
+        # The directory containing this PyArrow C++ CMakeLists.txt
+        source_pyarrow_cpp = pjoin(source, "pyarrow/src")
+
+        # The directory for the module being built
+        build_dir = pjoin(os.getcwd(), 'build', 'cpp')
+
+        # The directory containing Arrow C++ build
+        arrow_build_dir = os.environ.get('ARROW_BUILD_DIR', 'build')
+
+        if not os.path.isdir(build_dir):
+            self.mkpath(build_dir)
+
+        # Change to the build directory
+        with changed_dir(build_dir):
+            # cmake args
+            cmake_options = [
+                '-DARROW_BUILD_DIR=' + str(arrow_build_dir),
+                '-DCMAKE_BUILD_TYPE=' + str(self.build_type.lower()),
+                '-DCMAKE_INSTALL_LIBDIR=lib',
+                '-DCMAKE_INSTALL_PREFIX=' + str(pyarrow_cpp_home),
+                '-DPYTHON_EXECUTABLE=' + sys.executable,
+                '-DPython3_EXECUTABLE=' + sys.executable,
+                '-DPYARROW_CXXFLAGS=' + str(self.cmake_cxxflags),
+            ]
+
+            # Check for specific options
+            def append_cmake_bool(value, varname):
+                cmake_options.append('-D{0}={1}'.format(
+                    varname, 'on' if value else 'off'))
+
+            append_cmake_bool(self.with_dataset, 'PYARROW_WITH_DATASET')
+            append_cmake_bool(self.with_parquet_encryption,
+                              'PYARROW_WITH_PARQUET_ENCRYPTION')
+            append_cmake_bool(self.with_hdfs,
+                              'PYARROW_WITH_HDFS')
+
+            # Windows
+            if self.cmake_generator:
+                cmake_options += ['-G', self.cmake_generator]
+
+            # build args
+            build_tool_args = []
+            if os.environ.get('PYARROW_BUILD_VERBOSE', '0') == '1':
+                cmake_options.append('-DCMAKE_VERBOSE_MAKEFILE=ON')
+            if os.environ.get('PYARROW_PARALLEL'):
+                build_tool_args.append('--')
+                build_tool_args.append(
+                    '-j{0}'.format(os.environ['PYARROW_PARALLEL']))
+
+            # run cmake
+            print("-- Running CMake for PyArrow C++")
+            self.spawn(['cmake'] + cmake_options + [source_pyarrow_cpp])
+            print("-- Finished CMake for PyArrow C++")
+            # run make & install
+            print("-- Running CMake build and install for PyArrow C++")
+            self.spawn(['cmake', '--build', '.', '--config',
+                       self.build_type, '--target', 'install'] +
+                       build_tool_args)
+            print("-- Finished CMake build and install for PyArrow C++")
+
+    def _run_cmake(self, pyarrow_cpp_home):
+        # check if build_type is correctly passed / set
+        if self.build_type.lower() not in ('release', 'debug',
+                                           'relwithdebinfo'):
+            raise ValueError("--build-type (or PYARROW_BUILD_TYPE) needs to "
+                             "be 'release', 'debug' or 'relwithdebinfo'")
 
         # The directory containing this setup.py
         source = os.path.dirname(os.path.abspath(__file__))
 
         # The staging directory for the module being built
         build_cmd = self.get_finalized_command('build')
-        build_temp = pjoin(os.getcwd(), build_cmd.build_temp)
-        build_lib = pjoin(os.getcwd(), build_cmd.build_lib)
         saved_cwd = os.getcwd()
+        build_temp = pjoin(saved_cwd, build_cmd.build_temp)
+        build_lib = pjoin(saved_cwd, build_cmd.build_lib)
 
         if not os.path.isdir(build_temp):
             self.mkpath(build_temp)
 
+        if self.inplace:
+            # a bit hacky
+            build_lib = saved_cwd
+
         # Change to the build directory
         with changed_dir(build_temp):
             # Detect if we built elsewhere
@@ -235,13 +331,20 @@ class build_ext(_build_ext):
                                      cachefile.read()).group(1)
                 cachefile.close()
                 if (cachedir != build_temp):
+                    build_base = pjoin(saved_cwd, build_cmd.build_base)
+                    print(f"-- Skipping build. Temp build {build_temp} does "
+                          f"not match cached dir {cachedir}")
+                    print("---- For a clean build you might want to delete "
+                          f"{build_base}.")
                     return
 
             static_lib_option = ''
 
             cmake_options = [
-                '-DPYTHON_EXECUTABLE=%s' % sys.executable,
-                '-DPython3_EXECUTABLE=%s' % sys.executable,
+                '-DPYTHON_EXECUTABLE=' + sys.executable,
+                '-DPython3_EXECUTABLE=' + sys.executable,
+                '-DPYARROW_CPP_HOME=' + str(pyarrow_cpp_home),
+                '-DPYARROW_CXXFLAGS=' + str(self.cmake_cxxflags),
                 static_lib_option,
             ]
 
@@ -253,12 +356,16 @@ class build_ext(_build_ext):
                 cmake_options += ['-G', self.cmake_generator]
 
             append_cmake_bool(self.with_cuda, 'PYARROW_BUILD_CUDA')
+            append_cmake_bool(self.with_substrait, 'PYARROW_BUILD_SUBSTRAIT')
             append_cmake_bool(self.with_flight, 'PYARROW_BUILD_FLIGHT')
             append_cmake_bool(self.with_gandiva, 'PYARROW_BUILD_GANDIVA')
             append_cmake_bool(self.with_dataset, 'PYARROW_BUILD_DATASET')
             append_cmake_bool(self.with_orc, 'PYARROW_BUILD_ORC')
             append_cmake_bool(self.with_parquet, 'PYARROW_BUILD_PARQUET')
+            append_cmake_bool(self.with_parquet_encryption,
+                              'PYARROW_BUILD_PARQUET_ENCRYPTION')
             append_cmake_bool(self.with_plasma, 'PYARROW_BUILD_PLASMA')
+            append_cmake_bool(self.with_gcs, 'PYARROW_BUILD_GCS')
             append_cmake_bool(self.with_s3, 'PYARROW_BUILD_S3')
             append_cmake_bool(self.with_hdfs, 'PYARROW_BUILD_HDFS')
             append_cmake_bool(self.with_tensorflow, 'PYARROW_USE_TENSORFLOW')
@@ -295,18 +402,14 @@ class build_ext(_build_ext):
                         '-j{0}'.format(os.environ['PYARROW_PARALLEL']))
 
             # Generate the build files
-            print("-- Running cmake for pyarrow")
+            print("-- Running cmake for PyArrow")
             self.spawn(['cmake'] + extra_cmake_args + cmake_options + [source])
-            print("-- Finished cmake for pyarrow")
+            print("-- Finished cmake for PyArrow")
 
-            print("-- Running cmake --build for pyarrow")
+            print("-- Running cmake --build for PyArrow")
             self.spawn(['cmake', '--build', '.', '--config', self.build_type] +
                        build_tool_args)
-            print("-- Finished cmake --build for pyarrow")
-
-            if self.inplace:
-                # a bit hacky
-                build_lib = saved_cwd
+            print("-- Finished cmake --build for PyArrow")
 
             # Move the libraries to the place expected by the Python build
             try:
@@ -314,17 +417,42 @@ class build_ext(_build_ext):
             except OSError:
                 pass
 
+            def copy_libs(dir):
+                for path in os.listdir(pjoin(pyarrow_cpp_home, dir)):
+                    if "python" in path:
+                        pyarrow_path = pjoin(build_lib, "pyarrow", path)
+                        if os.path.exists(pyarrow_path):
+                            os.remove(pyarrow_path)
+                        pyarrow_cpp_path = pjoin(pyarrow_cpp_home, dir, path)
+                        print(f"Copying {pyarrow_cpp_path} to {pyarrow_path}")
+                        shutil.copy(pyarrow_cpp_path, pyarrow_path)
+
+            # Move libraries to python/pyarrow
+            # For windows builds, move DLL from bin/
+            try:
+                copy_libs("bin")
+            except OSError:
+                pass
+            copy_libs("lib")
+
             if sys.platform == 'win32':
                 build_prefix = ''
             else:
                 build_prefix = self.build_type
 
             if self.bundle_arrow_cpp or self.bundle_arrow_cpp_headers:
-                print('Bundling includes: ' + pjoin(build_prefix, 'include'))
-                if os.path.exists(pjoin(build_lib, 'pyarrow', 'include')):
-                    shutil.rmtree(pjoin(build_lib, 'pyarrow', 'include'))
-                shutil.move(pjoin(build_prefix, 'include'),
-                            pjoin(build_lib, 'pyarrow'))
+                arrow_cpp_include = pjoin(build_prefix, 'include')
+                print('Bundling includes: ' + arrow_cpp_include)
+                pyarrow_include = pjoin(build_lib, 'pyarrow', 'include')
+                if os.path.exists(pyarrow_include):
+                    shutil.rmtree(pyarrow_include)
+                shutil.move(arrow_cpp_include, pyarrow_include)
+
+                # pyarrow/include file is first deleted in the previous step
+                # so we need to add the PyArrow C++ include folder again
+                pyarrow_cpp_include = pjoin(pyarrow_cpp_home, 'include')
+                shutil.move(pjoin(pyarrow_cpp_include, 'arrow', 'python'),
+                            pjoin(pyarrow_include, 'arrow', 'python'))
 
             # Move the built C-extension to the place expected by the Python
             # build
@@ -337,7 +465,7 @@ class build_ext(_build_ext):
                         print('Cython module {0} failure permitted'
                               .format(name))
                         continue
-                    raise RuntimeError('pyarrow C-extension failed to build:',
+                    raise RuntimeError('PyArrow C-extension failed to build:',
                                        os.path.abspath(built_path))
 
                 # The destination path to move the built C extension to
@@ -376,10 +504,10 @@ class build_ext(_build_ext):
         move_shared_libs(build_prefix, build_lib, "arrow_python")
         if self.with_cuda:
             move_shared_libs(build_prefix, build_lib, "arrow_cuda")
+        if self.with_substrait:
+            move_shared_libs(build_prefix, build_lib, "arrow_substrait")
         if self.with_flight:
             move_shared_libs(build_prefix, build_lib, "arrow_flight")
-            move_shared_libs(build_prefix, build_lib,
-                             "arrow_python_flight")
         if self.with_dataset:
             move_shared_libs(build_prefix, build_lib, "arrow_dataset")
         if self.with_plasma:
@@ -413,12 +541,18 @@ class build_ext(_build_ext):
     def _failure_permitted(self, name):
         if name == '_parquet' and not self.with_parquet:
             return True
+        if name == '_parquet_encryption' and not self.with_parquet_encryption:
+            return True
         if name == '_plasma' and not self.with_plasma:
             return True
         if name == '_orc' and not self.with_orc:
             return True
         if name == '_flight' and not self.with_flight:
             return True
+        if name == '_substrait' and not self.with_substrait:
+            return True
+        if name == '_gcsfs' and not self.with_gcs:
+            return True
         if name == '_s3fs' and not self.with_s3:
             return True
         if name == '_hdfs' and not self.with_hdfs:
@@ -429,6 +563,10 @@ class build_ext(_build_ext):
                 self.with_orc and self.with_dataset
         ):
             return True
+        if name == '_dataset_parquet' and not (
+                self.with_parquet and self.with_dataset
+        ):
+            return True
         if name == '_cuda' and not self.with_cuda:
             return True
         if name == 'gandiva' and not self.with_gandiva:
@@ -527,7 +665,7 @@ def _move_shared_libs_unix(build_prefix, build_lib, lib_name):
 
 # If the event of not running from a git clone (e.g. from a git archive
 # or a Python sdist), see if we can set the version number ourselves
-default_version = '6.0.0'
+default_version = '10.0.0'
 if (not os.path.exists('../.git') and
         not os.environ.get('SETUPTOOLS_SCM_PRETEND_VERSION')):
     os.environ['SETUPTOOLS_SCM_PRETEND_VERSION'] = \
@@ -546,7 +684,7 @@ def parse_git(root, **kwargs):
     """
     from setuptools_scm.git import parse
     kwargs['describe_command'] =\
-        'git describe --dirty --tags --long --match "apache-arrow-[0-9].*"'
+        'git describe --dirty --tags --long --match "apache-arrow-[0-9]*.*"'
     return parse(root, **kwargs)
 
 
@@ -581,9 +719,15 @@ else:
 
 
 if strtobool(os.environ.get('PYARROW_INSTALL_TESTS', '1')):
-    packages = ['pyarrow', 'pyarrow.tests']
+    packages = find_namespace_packages(include=['pyarrow*'])
+    exclude_package_data = {}
 else:
-    packages = ['pyarrow']
+    packages = find_namespace_packages(include=['pyarrow*'],
+                                       exclude=["pyarrow.tests*"])
+    # setuptools adds back importable packages even when excluded.
+    # https://github.com/pypa/setuptools/issues/3260
+    # https://github.com/pypa/setuptools/issues/3340#issuecomment-1219383976
+    exclude_package_data = {"pyarrow": ["tests*"]}
 
 
 setup(
@@ -592,6 +736,7 @@ setup(
     zip_safe=False,
     package_data={'pyarrow': ['*.pxd', '*.pyx', 'includes/*.pxd']},
     include_package_data=True,
+    exclude_package_data=exclude_package_data,
     distclass=BinaryDistribution,
     # Dummy extension to trigger build_ext
     ext_modules=[Extension('__dummy__', sources=[])],
@@ -613,20 +758,24 @@ setup(
     setup_requires=['setuptools_scm', 'cython >= 0.29'] + setup_requires,
     install_requires=install_requires,
     tests_require=['pytest', 'pandas', 'hypothesis'],
-    python_requires='>=3.6',
+    python_requires='>=3.7',
     description='Python library for Apache Arrow',
     long_description=long_description,
     long_description_content_type='text/markdown',
     classifiers=[
         'License :: OSI Approved :: Apache Software License',
-        'Programming Language :: Python :: 3.6',
         'Programming Language :: Python :: 3.7',
         'Programming Language :: Python :: 3.8',
         'Programming Language :: Python :: 3.9',
+        'Programming Language :: Python :: 3.10',
     ],
     license='Apache License, Version 2.0',
     maintainer='Apache Arrow Developers',
     maintainer_email='dev@arrow.apache.org',
     test_suite='pyarrow.tests',
-    url='https://arrow.apache.org/'
+    url='https://arrow.apache.org/',
+    project_urls={
+        'Documentation': 'https://arrow.apache.org/docs/python',
+        'Source': 'https://github.com/apache/arrow',
+    },
 )
